unit USerialNGXE;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

// DomIS Internet Solutions http://www.domis.de
// Visit SerialNG Homepage http://www.domis.de/serialng.htm

// This Source is distributed under the terms of Open-Source
// This mean You can use this Source free for Open-Source development
// Additionally I allow the use for any inhouse Projects
// If You want to make a Closed-Source Project with this Source,
// You have to reference Back to the Source and have to distribute the Source
// Any changes should be marked and this header should remain here
// Under all circumstances it is prohibited to use this source for military Products!!!
// Refer the readme.txt

// This is Version 3 of the Basic Communication Component
// it is a complete revision of SerialNG
// to be used in Lazarus FPC (http://www.lazarus-ide.org/)
// It should support 32 and 64 bit

// Thank You for using and supporting SerialNG!

// Installation

// Usage
// Please take a look to the Demofiles.
// Start with SerialXEBasicDemo.lpi, this contains a very simple approach to the component
// without the need to install the component to the FPC Component Repository

// The Base of the Version 1.X of this unit is taken from "TSerialPort: Basic Serial Communications in Delphi"
// created by Jason "Wedge" Perry, but I could not find him again

// PC serial port Pins are as follows
// Name Dir  9Pin 25Pin
// DCD   In   1    8
// RXD   In   2    3
// TXD   Out  3    2
// DTR   Out  4    20
// GND   -    5    7
// DSR   In   6    6
// RTS   Out  7    4
// CTS   In   8    5
// RI    In   9    22
// Dir means the direction from the ports view (e.g. DCD is an input, You can read this port)

// Version History
// All Version are available at http://www.domis.de/serialng.htm
// 3.0.1.1 May 2017  New DelphiXE Version
// 3.0.1.2 Jan 2018  Added DosQueryDeviceA to get some more strange COM-Port in the list
// 3.0.1.3 Feb 2018  Added '\\.\' in front of CommPortname to access Ports COM10 and higher
// 3.0.1.4 May 2018  First transformation to Lazarus (IDE 1.8.4, FPC 3.0.4)
interface
{$BOOLEVAL OFF}
uses
{$IFnDEF FPC}
  System.Classes, System.SysUtils, Winapi.Messages, Winapi.Windows, System.SyncObjs, System.Types, System.Contnrs, System.Win.Registry,
{$ELSE}
  Classes, Windows, SyncObjs, Contnrs, Registry,
{$ENDIF}
  SysUtils;

type
  TSNXParityType = (ptNoParity,ptOddParity,ptEvenParity,ptMarkParity,ptSpaceParity);
  TSNXStopBits = (sbOneStopBit,sbOnePtFiveStopBits,sbTwoStopBits);
  TSNXFlowControlFlag = (fcOutxCtsFlow, // CTS output flow control
                         fcOutxDsrFlow, // DSR output flow control
                         fcDtrControlEnable,     // DTR Enable
                         fcDtrControlHandshake,  // DTR Handshake
                         fcDsrSensitivity,       // DSR sensitivity
                         fcTXContinueOnXoff,   // XOFF continues Tx
                         fcOutX,               // XON/XOFF out flow control
                         fcInX,                // XON/XOFF in flow control
                         fcRtsControlEnable,    // RTS Enable
                         fcRtsControlHandshake, // RTS Enable
                         fcRtsControlToggle);    // RTS Enable
  TSNXFlowControlFlags = set of TSNXFlowControlFlag;
// Definitions for the DCB found in windows.pas for reference only
const
  BaudRateCount = 15;
  BaudRates : array[0..BaudRateCount-1] of Cardinal =
    (CBR_110, CBR_300, CBR_600, CBR_1200, CBR_2400, CBR_4800, CBR_9600, CBR_14400,
     CBR_19200, CBR_38400, CBR_56000, CBR_57600, CBR_115200, CBR_128000, CBR_256000);
// The business with the TimeOuts during sending Data, and esspeccialy during
// reception of Data depends on the selected Baudrate.
// while on 9600 Baud every 1ms a Char is received, is this time on 256KBaud only 4µs
// Strange enough, windows support only 1ms as shortest Intervall!
// Below some standard TimeOuts for the given Baudrates in µs
  XTOCharDelayDef : array [0..BaudRateCount-1] of Cardinal =
    (100000, 37000, 18000, 9000, 4500, 2300, 1100, 760, 570, 290, 200, 190, 95, 85, 43);

const
// Some basic FlowControlModes
  fcNone : TSNXFlowControlFlags = [];
  fcXONXOFF : TSNXFlowControlFlags = [fcOutX, fcInX, fcTXContinueOnXoff];
  fcRTSCTS : TSNXFlowControlFlags = [fcOutxCtsFlow, fcRtsControlHandshake];
  fcDSRDTR : TSNXFlowControlFlags = [fcOutxDsrFlow, fcDtrControlHandshake];

// Constants for using in the ErrorNoise Property
// I tried to catch Errors, send Warnings or sometimes Messages for Your convinience.
// If You set the ErrorNoise Property to one of the Values this kind of Messages will be reported
// while those with higher Numbers not
const
  enError = 0;   // Errors like unable to Open Port
  enWarning = 1; // Warnings like cant pause Thread
  enMsg = 2;     // Messages like starting with Port opening
  enDebug = 3;   // Debughelpermessage like im am here or there
  enAll = 255;   // Show all

const
// Set some constant defaults.
// These are the equivalent of COM2:9600,N,8,1;
  dfltCommPort = 'COM2';
  dfltBaudRate = CBR_9600;
  dfltParityType = ptNoParity; //NOPARITY;
  dfltParityErrorChar = #0;
  dfltParityErrorReplacement = False;
  dfltStopBits = sbOneStopBit; //ONESTOPBIT;
  dfltDataBits = 8;
  dfltXONChar = #$11;  {ASCII 11h}
  dfltXOFFChar = #$13; {ASCII 13h}
  dfltXONLimDiv = 33; // Set the XONLimit to dflt_XONLimDiv/100*RxQueueSize
  dfltXOFFLimDiv = 33; // Set the XONLimit to dflt_XONLimDiv/100*RxQueueSize
  dfltFlowControlFlags = []; //No FlowControl
  dfltStripNullChars = False;
  dfltBinaryMode = True;
  dfltEventChar = #0;
// Timeout defaults, fits only to 9600 Baud!
// Look that CharDelay is in micro, ExtraDely in milli Seconds!
  dfltRTOCharDelayTime = 1100; // 1100microsec (1100E-06 secs)
  dfltRTOExtraDelayTime = 250; // 250msec      ( 250E-03 secs)
  dfltWTOCharDelayTime = 1100; // 1100microsec (1100E-06 secs)
  dfltWTOExtraDelayTime = 250; // 1000msec     ( 250E-03 secs)
  dfltXTOAuto = True; // lets the Component adjust CharDelay timing on every Baudrate change
  dfltClusterSize = 3072; //Max Clustersize
  dfltRTSState = True;
  dfltDTRState = True;
  dfltErrorNoise = enMsg;
  dfltBREAKState = False;
  dfltRxQueueSize = 4096;
  dfltTxQueueSize = 4096;
  dfltThreadQuietMode = False;  // Thread will call On xxx Events
  dfltUseThreadSynchronize = True; //Thread will use Synchronize Method for calling Events
  dfltAutoReadRequest = False;
type

// Special Event function for the Error and Warning
  TNotifyErrorEvent = procedure(Sender : TObject;Place, Code: Cardinal; Msg : AnsiString; Noise : Byte) of object;

// TSerialCluster is a Object for the Receiving Process.
// If You just want to receive some Data, don't care
// Under normal circumstances You have not to deal with this Object

// I decided to realize the receiving Process as follow:
// Between two cycles in the WorkThread may the SerialPort receive more than one Character.
// Basicly those Chars are named "Cluster" (and not "Telegram", as You might expect)
// How many Chars are stored in one TSerialCluster depeds on, User controlled,  Conditions
// 1. the ClusterSize is reached
// 2. the Receivingtime is reached
// 3. an (Line-) Error occoured
// 4. The Program set the "ReadRequest" Property to "True"
// If the condition is met, the received Chars are Stored as a Cluster into the ClusterList

  TSerialNGXEDataCluster = class (TObject)
  private
    FCreationTime : TDateTime;
    FCreationTick : Cardinal;
    FClusterData : array of Byte; // Datafield
    FClusterSize : Integer; // How many Bytes in Datafield
    FClusterCCError : Cardinal; // This Value comes from the ClearCommError function and is a Bitfield
      // CE_RXOVER       = 1 Receive Queue overflow
      // CE_OVERRUN      = 2 Receive Overrun Error
      // CE_RXPARITY     = 4 Receive Parity Error
      // CE_FRAME        = 8 Receive Framing error
      // CE_BREAK      = $10 Break Detected
      // CE_TXFULL    = $100 TX Queue is full
      // CE_MODE     = $8000 Requested mode unsupported
    function GetClusterSize : Integer;
    function GetClusterData(Index : Integer) : Byte;
  public
    property CreationTime : TDateTime read FCreationTime;
    property CreationTick : Cardinal read FCreationTick;
    property ClusterSize : Integer read GetClusterSize;
    property ClusterCCError : Cardinal read FClusterCCError;
    property ClusterData[Index : Integer] : Byte read GetClusterData; //Slow access, better use one of the Methods below!
    function GetDataAsAnsiString : AnsiString; //Copy of all Data in an AnsiString
    function GetDataAsString : String; //Same as: String(GetDataAsAnsiString)
    function GetDataAsUnicodeString : UnicodeString;
    procedure CopyClusterData(var Destination); //Copy of the buffer without any preprocessing
    constructor Create(const AData; const ASize : Integer; const ACCError : Cardinal);
    destructor Destroy; override;
  end;

  TSerialPortNGXEWorkThread = class; // A kind of forward declaration, for the worker thread

// TSerialPortNGXE is the Heart of the Job, everything what has to do with the
// SerialPort should be done with this Component
// The Concept is as follows:
// After instancing the Port is still closed
// You should set the Property "CommPort" to the full Name of the Port e.g. 'COM1'
// Set other configuration parameters a desired
// Then set the "Active" Property to "True", verify that the "Active" property has been set to True (otherwise check error)
// Sending Data ist performed with the SendData or the SendString procedures
// since Sendig is "Overlapped" the Method returns immidiatly and You can do
// some other Jobs in Your main Programm
// You should not reentry this Procedures until they done there Job (Will give a warning).
// If send is done the component call the "OnWriteDone" Event.
// You also can ask the "SendInProgress" Property.
// Reading Data is fairly simple, just Read the Data with one of the "ReadNextCluster..." functions
// You place the read Access into the "OnRxClusterEvent".
// See sample Programs for more Information
  TSerialPortNGXE = class(TComponent)
  private
    // Variables holding Values for Properties
    FCommPort : String;
    FBaudRate : Cardinal;
    FParityType : TSNXParityType;
    FParityErrorChar : AnsiChar;
    FParityErrorReplacement : Boolean;
    FStopBits : TSNXStopBits;
    FDataBits : Byte;
    FXONChar : AnsiChar;
    FXOFFChar : AnsiChar;
    FXONLimDiv : Byte; // 0..100
    FXOFFLimDiv : Byte; // 0..100
    FFlowControlFlags : TSNXFlowControlFlags;
    FStripNullChars : Boolean;  // Strip null chars?
    FEventChar : AnsiChar;
    FErrorNoise : Byte;
    // These fields are set in the EventThread
    FCommStateFlags : Cardinal;
    FCommStateInQueue : Cardinal;
    FCommStateOutQueue : Cardinal;
    FCommError : Cardinal;
    FCommEvent : Cardinal;
    FModemState : Cardinal;
    // TimeOut definitions
    FRTOCharDelayTime : Cardinal;     // in µs max: 4.294.967.295µs aprox 1h 20min
    FRTOExtraDelayTime : Word;    // in ms
    FWTOCharDelayTime : Cardinal;     // in µs
    FWTOExtraDelayTime : Word;    // in ms
    FXTOAuto : Boolean;
    FActive : Boolean;
    FRTSState : Boolean;
    FDTRState : Boolean;
    FBREAKState : Boolean;
    FCTSState : Boolean;
    FDSRState : Boolean;
    FRLSDState : Boolean;
    FRingState : Boolean;
    FClusterSize : Word;
    FRxQueueSize : Word;
    FTxQueueSize : Word;
    FReadRequest : Boolean; // Force Thread to Read the Queue
    FSendInProgress : Boolean;
    FWrittenBytes : Cardinal;
    FThreadQuietMode : Boolean;
    FUseThreadSynchronize : Boolean;
    FAutoReadRequest : Boolean;
    // Eventvariables
    FOnTxQueueEmptyEvent : TNotifyEvent;
    FOnCommEvent : TNotifyEvent;
    FOnCommStat : TNotifyEvent;
    FOnBreakEvent : TNotifyEvent;
    FOnCTSEvent : TNotifyEvent;
    FOnDSREvent : TNotifyEvent;
    FOnLineErrorEvent : TNotifyEvent;
    FOnRingEvent : TNotifyEvent;
    FOnRIEvent : TNotifyEvent;
    FOnRLSDEvent : TNotifyEvent;
    FOnRxClusterEvent : TNotifyEvent;
    FOnRxCharEvent : TNotifyEvent;
    FOnRxEventCharEvent : TNotifyEvent;
    FOnWriteDone : TNotifyEvent;
    FOnProcessError : TNotifyErrorEvent;

    FHCommPort : THandle; // Handle to the port.
    WriteOverlapped : TOverlapped; //Overlapped field for Write
    ReadOverlapped : TOverlapped;  //Overlapped field for Read
    StatusOverlapped : TOverlapped; //Overlapped field for Status
    BytesToWrite : Cardinal;
    WriteStartTime : Cardinal;
    WorkThread : TSerialPortNGXEWorkThread;
    WorkThreadIsRunning : Boolean;
    WorkThreadIsTerminated : Boolean;
    ShutdownInProgress : Boolean;
    FReceiveDataClusterList : TObjectList;
    FReceiveDataClusterListLockCS : TCriticalSection;
    LastErr : Integer;
    WinVersion : Integer; // 0 Win32s on Win3.11, 1 Win 9x, 2 WinNT
    // Procedures for setting the variables, refrenced in the Properties
    procedure SetCommPort(Value : String);
    procedure SetBaudRate(Value : Cardinal);
    procedure SetParityType(Value : TSNXParityType);
    procedure SetParityErrorChar(Value : AnsiChar);
    procedure SetParityErrorReplacement(Value : Boolean);
    procedure SetStopBits(Value : TSNXStopBits);
    procedure SetDataBits(Value : Byte);
    procedure SetXONChar(Value : AnsiChar);
    procedure SetXOFFChar(Value : AnsiChar);
    procedure SetXONLimDiv(Value : Byte);
    procedure SetXOFFLimDiv(Value : Byte);
    procedure SetFlowControlFlags(Value : TSNXFlowControlFlags);
    procedure SetStripNullChars(Value : Boolean);
    procedure SetEventChar(Value : AnsiChar);
    procedure SetRTOCharDelayTime(Value : Cardinal);
    procedure SetRTOExtraDelayTime(Value : Word);
    procedure SetWTOCharDelayTime(Value : Cardinal);
    procedure SetWTOExtraDelayTime(Value : Word);
    procedure SetXTOAuto(Value : Boolean);
    procedure SetClusterSize(Value : Word);
    procedure SetRxQueueSize(Value : Word);
    procedure SetTxQueueSize(Value : Word);
    procedure SetErrorNoise(Value : Byte);
    procedure SetSignalRTS(State : Boolean);
    procedure SetSignalDTR(State : Boolean);
    procedure SetSignalBREAK(State : Boolean);
    procedure SetReadRequest(Value : Boolean);
    procedure SetActive(NewState : Boolean);
    // Rest of Procedures
    procedure InitOverlapped(var Overlapped : TOverlapped);
    procedure ResetOverlapped(var Overlapped : TOverlapped);
    procedure SetOverlapped(var Overlapped : TOverlapped);
    procedure SetupDCB;
    procedure PortWork(const ReOpen : Boolean); //If ReOpen is True the Port will be (Re-) Opened, otherwise closed. The ActiveFlag will bes Set!
    procedure EnableEvents;
    procedure ProcessError(const Place, Code : Cardinal; const Msg : AnsiString; const Noise : Byte);
    procedure WaitForThreadNotRunning(const WaitMS : Integer = 1000); //WaitMS waittime in Milliseconds
  published
    //If You set Active to True, the component tries to Open the Port, if Opened the state remains True.
    property Active : Boolean read FActive write SetActive default False;
    property ComHandle : THandle read FHCommPort; // default INVALID_HANDLE_VALUE;
    property CommPort : String read FCommPort write SetCommPort;
    property BaudRate :  Cardinal read FBaudRate write SetBaudRate default dfltBaudRate;
    property ParityType : TSNXParityType read FParityType write SetParityType default dfltParityType;
    property ParityErrorChar : AnsiChar read fParityErrorChar write SetParityErrorChar default dfltParityErrorChar;
    property ParityErrorReplacement :  Boolean read fParityErrorReplacement  write SetParityErrorReplacement  default dfltParityErrorReplacement;
    property StopBits : TSNXStopBits read FStopBits write SetStopBits default dfltStopBits;
    // Supported Databits are 5,6,7,8 but not all combinations with StopBits are allowed
    // additionally some allowed combinations may not be supported by all modems.
    // E.g 8 Bit, Even Parity, One StopBit may not be usabel in several Modems
    // If using less than 8 Bits, Strings may ruined because bits containing information are not transmitted
    property DataBits : Byte read FDataBits write SetDataBits default dfltDataBits;
    property XONChar : AnsiChar read FXONChar  write SetXONChar  default dfltXONChar;
    property XOFFChar : AnsiChar read FXOFFChar write SetXOFFChar  default dfltXOFFChar;
    property XONLimDiv : Byte read FXONLimDiv  write SetXONLimDiv default dfltXOnLimDiv;
    property XOFFLimDiv : Byte read FXOFFLimDiv  write SetXOFFLimDiv default dfltXOffLimDiv;
    property FlowControlFlags : TSNXFlowControlFlags read FFlowControlFlags write SetFlowControlFlags default dfltFlowControlFlags;
    property StripNullChars : Boolean read FStripNullChars write SetStripNullChars default dfltStripNullChars;
    property EventChar : AnsiChar read FEventChar write SetEventChar default dfltEventChar;

// One part of the Clusterdefinition is here, please read carefully
// The "RTOCharDelayTime" is the Time that may delay between two received Chars
// This Time should be computed depending from the Baudrate e.g. 9600 Baud -> 960 Chars per Second -> Delay 1ms
// You can use the CharDelayDefault Procedure to set RTOCharDelayTime and WTOCharDelayTime depending
// of the selected Baudrate!
    property RTOCharDelayTime : Cardinal read FRTOCharDelayTime write SetRTOCharDelayTime default dfltRTOCharDelayTime;
// The "RTOExtraDelayTime" is the Time that may delay addionally once
// So if the (CharCount*RTOCharDelayTime)/1000 + RTOExtraDelayTime extends the measured Time, a Cluster will be build
    property RTOExtraDelayTime : Word read FRTOExtraDelayTime write SetRTOExtraDelayTime default dfltRTOExtraDelayTime;
// Clustersize specify how long one Cluster could become max
    property ClusterSize : Word read FClusterSize write SetClusterSize default dfltClusterSize;
// RxQueueSize and TxQueueSize specify the amount of chars that could be received or transmitted in once
// without interaction of the program.
// The values should be longer than the Cluster size to prevent overrun errors
// The values are used in the device driver
    property RxQueueSize : Word read FRxQueueSize write SetRxQueueSize default dfltRxQueueSize;
    property TxQueueSize : Word read FTxQueueSize write SetTxQueueSize default dfltTxQueueSize;

    property WTOCharDelayTime : Cardinal read FWTOCharDelayTime write SetWTOCharDelayTime default dfltWTOCharDelayTime;
    property WTOExtraDelayTime : Word read FWTOExtraDelayTime write SetWTOExtraDelayTime default dfltWTOExtraDelayTime;
    property XTOAuto : Boolean read FXTOAuto write SetXTOAuto default dfltXTOAuto;
    property RTSState : Boolean read FRTSState write SetSignalRTS default dfltRTSState;
    property DTRState : Boolean read FDTRState write SetSignalDTR default dfltDTRState;
    property BREAKState : Boolean read FBREAKState write SetSignalBREAK default dfltBreakState;
    property CTSState : Boolean read FCTSState;
    property DSRState : Boolean read FDSRSTate;
    property RLSDState : Boolean read FRLSDState;
    property RingState : Boolean read FRingState;
    property ErrorNoise : Byte read FErrorNoise write SetErrorNoise default dfltErrorNoise;
    property ReadRequest : Boolean read FReadRequest write SetReadRequest default False;
    property SendInProgress : Boolean read FSendInProgress;
    property CommError : Cardinal read FCommError;
    property CommStateFlags : Cardinal read FCommStateFlags;
    property CommStateInQueue: Cardinal read FCommStateInQueue;
    property CommStateOutQueue : Cardinal read FCommStateOutQueue;
    property ModemState : Cardinal read FModemState;
    property CommEvent : Cardinal read FCommEvent;
    property WrittenBytes : Cardinal read FWrittenBytes;
    // if ThreadQuietMode is set to True, no Events from the Thread are generated. Subsequently you have to implement
    // a different way of checking incoming data.
    // Rarely used!
    property ThreadQuietMode : Boolean read FThreadQuietMode write FThreadQuietMode default dfltThreadQuietMode;
    // if UseThreadSynchronize is set to False, the thread will call the Event callbacks directly
    // this is a very fast solution, which will also work in DLL use,
    // BUT you have to be very careful, since the code is running in the thread, not in the Window level
    property UseThreadSynchronize : Boolean read FUseThreadSynchronize write FUseThreadSynchronize default dfltUseThreadSynchronize;

    property AutoReadRequest : Boolean read FAutoReadRequest write FAutoReadRequest default dfltAutoReadRequest;

    // Event Properties
    property OnCommEvent : TNotifyEvent read FONCommEvent write FONCommEvent;
    property OnCommStat : TNotifyEvent read FONCommStat write FONCommStat;
    property OnTxQueueEmptyEvent : TNotifyEvent read FONTxQueueEmptyEvent write FONTxQueueEmptyEvent;
    property OnWriteDone : TNotifyEvent read FONWriteDone write FONWriteDone;
    property OnBreakEvent : TNotifyEvent read FONBreakEvent write FONBreakEvent;
    property OnCTSEvent : TNotifyEvent read FONCTSEvent write FONCTSEvent;
    property OnDSREvent : TNotifyEvent read FONDSREvent write FONDSREvent;
    property OnLineErrorEvent : TNotifyEvent read FONLineErrorEvent write FONLineErrorEvent;
    property OnRingEvent : TNotifyEvent read FONRingEvent write FONRingEvent; // RING, RING on falling edge of the RI Pin
    property OnRIEvent : TNotifyEvent read FONRIEvent write FONRIEvent;       // on every change of the RI Pin
    property OnRLSDEvent : TNotifyEvent read FONRLSDEvent write FONRLSDEvent;
    property OnRxClusterEvent : TNotifyEvent read FONRxClusterEvent write FONRxClusterEvent;
    property OnRxCharEvent : TNotifyEvent read FONRxCharEvent write FONRxCharEvent;
    property OnRxEventCharEvent : TNotifyEvent read FONRxEventCharEvent write FONRxEventCharEvent;
    property OnProcessError : TNotifyErrorEvent read FONProcessError write FONProcessError;
  public
    // Procedures for external calling
    constructor Create(AOwner : TComponent); override; //Create the Component
    destructor Destroy; override; //Destroy
    procedure SendData(const Data; const Size : Cardinal); //Send binary Data
    procedure SendAnsiString(const S : AnsiString); //Send String Data
    procedure SendString(const S : String); //Same as: SendAnsiString(AnsiString(AString));
    procedure SendUnicodeString(const S : UnicodeString); //Send a Unicode String
    // Clusterfunctions works on received Datapackages
    //Method ExtractNextCluster
    //Generic reception function. Caller has to free the passed Object.
    //If Nil is returned if no Data was received.
    function ExtractNextCluster : TSerialNGXEDataCluster;
    // Method NextClusterExists, returns True if a Cluster was received
    function NextClusterExists : Boolean;
    function GetNextClusterSize : Integer;
    function GetNextClusterCCError : Cardinal;
    //
    function CopyNextClusterData(var Data; const MaxLength : Integer) : Integer;
    function ReadNextClusterAsAnsiString : AnsiString;
    function ReadNextClusterAsString : String;
    function ReadNextClusterAsUnicodeString : UnicodeString;
    // Clears the Queues
    procedure ClearTxDQueue;
    procedure ClearRxDQueue;
    // Sets the Timingfields in depedecy to the Baudrate
    procedure AdjustXTODefaultValues;
    // Save and retrieves the Settings from a StringList
    procedure GetSettingsAsStrings(const Strings : TStrings);   //All current Properties will be written into Strings in a Form <propertyname>=<value>
    procedure WriteSettingsAsStrings(const Strings : TStrings); //All Properties will be set from Strings
  end;

// The TWorkThread class deals with several CommEvents and controll the receiving
// of Clusters and check the Sendprocess
// Under normal cirumstances You don't have to deal with
  TSerialPortNGXEWorkThread = class(TThread)
  private
    FOwner : TSerialPortNGXE;
    Place, Code : Cardinal;
    Msg : AnsiString;
    Noise : Byte;
    procedure ThreadSynchronize(Method: TThreadMethod);
    procedure SetProcessError(const APlace, ACode : Cardinal; const AMsg : AnsiString; const ANoise : Byte);
    procedure ProcessError;
    procedure RxClusterEvent;
    procedure CommEvent;
    procedure CommStatEvent;
    procedure BreakEvent;
    procedure CTSEvent;
    procedure DSREvent;
    procedure LineErrorEvent;
    procedure RingEvent;
    procedure RIEvent;
    procedure RLSDEvent;
    procedure RxCharEvent;
    procedure RxEventCharEvent;
    procedure TxQueueEmptyEvent;
    procedure WriteDone;
  public
    constructor Create(AOwner : TSerialPortNGXE);
    procedure Execute; override;
  end;

procedure Register;
procedure GetCommPortList(const Strings : TStrings);

implementation

procedure Register;
begin
  RegisterComponents('SerialNG', [TSerialPortNGXE]);
end;

// Bitmasks for the "Flags" Field of the DCB record
const
  bmfBinary              = $0001; // binary mode, no EOF check
  bmfParity              = $0002; // enable parity checking
  bmfOutxCtsFlow         = $0004; // CTS output flow control
  bmfOutxDsrFlow         = $0008; // DSR output flow control
// DTR Control Flow Values  DTR_CONTROL_DISABLE = 0;  DTR_CONTROL_ENABLE = 1;  DTR_CONTROL_HANDSHAKE = 2;
  bmfDtrControlEnable    = $0010; // DTR Enable
  bmfDtrControlHandshake = $0020; // DTR Handshake
  bmfDsrSensitivity      = $0040; // DSR sensitivity
  bmfTXContinueOnXoff    = $0080; // XOFF continues Tx
  bmfOutX                = $0100; // XON/XOFF out flow control
  bmfInX                 = $0200; // XON/XOFF in flow control
  bmfErrorChar           = $0400; // enable error replacement
  bmfNull                = $0800; // enable null stripping
// RTS Control Flow Values  RTS_CONTROL_DISABLE = 0;  RTS_CONTROL_ENABLE = 1;  RTS_CONTROL_HANDSHAKE = 2;  RTS_CONTROL_TOGGLE = 3;
  bmfRtsControlEnable    = $1000; // RTS Enable
  bmfRtsControlHandshake = $2000; // RTS Enable
  bmfRtsControlToggle    = $3000; // RTS Enable
  bmfAbortOnError        = $4000; // abort reads/writes on error


function CheckOS(var VersionInfo : TOSVersionInfo) : Integer;
begin
{
    dwOSVersionInfoSize: DWORD;
    dwMajorVersion: DWORD;
    dwMinorVersion: DWORD;
    dwBuildNumber: DWORD;
    dwPlatformId: DWORD;
    szCSDVersion: array[0..127] of AnsiChar; // Maintenance string for PSS usage
}
  VersionInfo.dwOSVersionInfoSize := SizeOf(VersionInfo);
  if GetVersionEx(VersionInfo) then
    CheckOS := VersionInfo.dwPlatformId
  else
    CheckOS := -1;
end;



procedure GetCommPortList(const Strings : TStrings);
var
  Reg : TRegistry;

  procedure ScanCom1To255;
  var
    PortName : AnsiString;
    i : Integer;
    DevicePath : AnsiString;
    rc : Cardinal;
  begin
    SetLength(DevicePath,255);
    for i := 1 to 255 do
    begin
      PortName := AnsiString(Format('COM%d',[i]));
      rc := QueryDosDeviceA(PAnsiChar(PortName),PAnsiChar(DevicePath),Length(DevicePath));
      if (rc > 0) then
        if Strings.IndexOf(String(PortName)) < 0 then
          Strings.Add(String(PortName));
    end;
  end;

  procedure ScanRegHardware;
  var
    i : integer;
    PortName : string;
    lName : TStringList;
  begin
    if Reg.OpenKeyReadOnly('\hardware\devicemap\serialcomm') then
    begin
      lName := TStringList.Create;
      try
        Reg.GetValueNames(lName);
        for i := 0 to lName.Count - 1 do
        begin
          if Reg.GetDataType(lName.Strings[i]) = rdString then
          begin
            PortName := Reg.ReadString(lName.Strings[i]);
            if Strings.IndexOf(PortName) < 0 then
              Strings.Add(PortName);
          end
        end;
      finally
        if Assigned(lName) then
          lName.Free;
      end;
    end
  end;

  procedure ScanRegEnum(Key : String);
  // This Subprocedure recurses thru all keys below key
  var
    lKey : TStringList;
    lName : TStringList;
    i : Integer;
    Driver, PortName : String;
    PortSubClass : Byte;
  begin
    PortSubClass := 0;
    if not Reg.OpenKeyReadOnly(Key) then // Danke Andreas Schmidt!
      Exit;
    lName := TStringList.Create;
    try
      Reg.GetValueNames(lName);
      i := lName.IndexOf('class');
      if i >= 0 then
      begin
        if (Reg.GetDataType('class') = rdString) and
           ((LowerCase(Reg.ReadString('class')) = 'ports') or    //normal Serialports like COMx
            (LowerCase(Reg.ReadString('class')) = 'modem')) and  //some abnormal onboard Modems
           (Reg.GetDataType('driver') = rdString) and
           (Reg.GetDataType('portname') = rdString) then
        begin
          Driver := Reg.ReadString('driver');
          PortName := Reg.ReadString('portname');
          if Reg.OpenKeyReadOnly('\System\CurrentControlSet\Services\Class\'+ Driver) and
             (Reg.ReadBinaryData('PortSubClass',PortSubClass,1) = 1) and
             ((PortSubClass = 1) or    //Ports
              (PortSubClass = 2)) and  //Modems
             (Strings.IndexOf(PortName) < 0) then
               Strings.Add(PortName);
        end
      end;
    finally
      if Assigned(lName) then
        lName.Free;
    end;
    Reg.OpenKeyReadOnly(Key);
    if Reg.HasSubKeys then
    begin
      lKey := TStringList.Create;
      try
        Reg.GetKeyNames(LKey);
        for i := 0 to LKey.Count - 1 do
          begin
            ScanRegEnum(Key + '\' + LKey[i]);
          end;
      finally
        if Assigned(lKey) then
          lKey.Free;
      end;
    end;
  end;

var
  VersionInfo : TOSVersionInfo;
begin
  Strings.Clear;
  VersionInfo.dwMajorVersion:= 0;
  if CheckOS(VersionInfo) > 0 then
    if VersionInfo.dwMajorVersion >= 4 then
    begin
      Reg := TRegistry.Create;
      try
        Reg.RootKey := HKEY_LOCAL_MACHINE;
        if VersionInfo.dwMajorVersion > 4 then
        begin  //Win 2k, XP
          if VersionInfo.dwMinorVersion > 0 then
            ScanCom1To255; //XP and all later Versions can do this
          ScanRegHardware;
        end
        else if VersionInfo.dwMinorVersion > 0 then
          ScanRegEnum('\enum') //Win9x, mE
        else
          ScanRegHardware; //WinNT
      finally
        Reg.Free;
      end;
    end;
end;


//------------------------------------------------------------------------------
// TSerialCluster Component
// Is an internal Component, as an average user you will not have to deal with this class
//------------------------------------------------------------------------------

constructor TSerialNGXEDataCluster.Create(const AData; const ASize : Integer; const ACCError : Cardinal);
begin
  inherited Create;
  FCreationTick := GetTickCount64;
  FCreationTime := Now;
  SetLength(FClusterData,ASize);
  if ASize > 0 then
    Move(AData,FClusterData[0],ASize);
  FClusterSize := ASize; // Size of Data
  FClusterCCError := ACCError;
end;

function TSerialNGXEDataCluster.GetClusterSize : Integer;
begin
  Result := Length(FClusterData);
end;

function TSerialNGXEDataCluster.GetClusterData(Index : Integer) : Byte;
begin
  Result := FClusterData[Index];
end;

procedure TSerialNGXEDataCluster.CopyClusterData(var Destination); //Copy of the buffer without any preprocessing
var
  sz : Integer;
begin
  sz := ClusterSize;
  if sz > 0 then
    Move(FClusterData[0],Destination, sz);
end;
function TSerialNGXEDataCluster.GetDataAsAnsiString : AnsiString; //Copy of all Data in an AnsiString
var
  sz : Integer;
begin
  sz := ClusterSize;
  SetLength(Result,sz);
  if sz > 0 then
    Move(FClusterData[0],Result[1], sz);
end;
function TSerialNGXEDataCluster.GetDataAsString : String; //Copy of all Data in a two Byte String
begin
  Result := String(GetDataAsAnsiString);
end;
function TSerialNGXEDataCluster.GetDataAsUnicodeString : UnicodeString;
var
  sz : Integer;
begin
  sz := ClusterSize;
  if Odd(sz) then // Für ungerade Byteanzahl, Platz für ein abschließendes NullByte lassen
    SetLength(Result,(sz+1) div 2)
  else
    SetLength(Result,sz div 2);
  if sz > 0 then
    Move(FClusterData[0],Result[1], sz);
end;

destructor TSerialNGXEDataCluster.Destroy;
begin
  SetLength(FClusterData,0);
  inherited;
end;


//------------------------------------------------------------------------------
// Class TSerialPortNGXE
//------------------------------------------------------------------------------
//
// Serveral "Set..." procedure for the Property mapping

procedure TSerialPortNGXE.SetCommPort(Value : String);
begin
  if Value <> FCommPort then
  begin
    FCommPort := value;
    PortWork(FActive);
  end;
end;

procedure TSerialPortNGXE.SetBaudRate(Value : Cardinal);
begin
  if Value <> FBaudRate then
  begin
    FBaudRate := Value;
    if FXTOAuto then
      AdjustXTODefaultValues; // Adjust the CharDelay Timeouts
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetParityType(Value : TSNXParityType);
begin
  if Value <> FParityType then
  begin
    FParityType := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetParityErrorChar(Value : AnsiChar);
begin
  if Value <> FParityErrorChar then
  begin
    FParityErrorChar := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetParityErrorReplacement(Value : Boolean);
begin
  if Value <> FParityErrorReplacement then
  begin
    FParityErrorReplacement := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetStopBits(Value : TSNXStopBits);
begin
  if Value <> FStopBits then
  begin
    FStopBits := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetDataBits(Value : Byte);
begin
  if Value <> FDataBits then
  begin
    FDataBits := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetXONChar(Value : AnsiChar);
begin
  if Value <> FXONChar then
  begin
    FXONChar := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetXOFFChar(Value : AnsiChar);
begin
  if Value <> FXOFFChar then
  begin
    FXOFFChar := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetXONLimDiv(Value : Byte);
begin
  if Value <> FXONLimDiv then
  begin
    if Value > 100 then
    begin
      ProcessError(0100,value,'Warning XOnLimDef set to 100',enWarning);
      value := 100;
    end;
    FXONLimDiv := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetXOFFLimDiv(Value : Byte);
begin
  if Value <> FXOFFLimDiv then
  begin
    if Value > 100 then
      begin
        ProcessError(0100,value,'Warning XOffLimDef set to 100',enWarning);
        Value := 100;
      end;
    FXOFFLimDiv := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetFlowControlFlags(Value : TSNXFlowControlFlags);
begin
  if Value <> FFlowControlFlags then
  begin
    FFlowControlFlags := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetStripNullChars(Value : Boolean);
begin
  if Value <> FStripNullChars then
  begin
    FStripNullChars := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetEventChar(Value : AnsiChar);
begin
  if Value <> FEventChar then
  begin
    FEventChar := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetRTOCharDelayTime(Value : Cardinal);
begin
  if Value <> FRTOCharDelayTime then
    FRTOCharDelayTime := Value;
end;

procedure TSerialPortNGXE.SetRTOExtraDelayTime(Value : Word);
begin
  if Value <> FRTOExtraDelayTime then
    FRTOExtraDelayTime := Value;
end;

procedure TSerialPortNGXE.SetWTOCharDelayTime(Value : Cardinal);
begin
  if Value <> FWTOCharDelayTime then
  begin
    FWTOCharDelayTime := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetWTOExtraDelayTime(Value : Word);
begin
  if Value <> FWTOExtraDelayTime then
  begin
    FWTOExtraDelayTime := Value;
    if FActive then
      SetupDCB;
  end;
end;

procedure TSerialPortNGXE.SetXTOAuto(Value : Boolean);
begin
  if Value <> FXTOAuto then
  begin
    FXTOAuto := Value;
    if FXTOAuto then
      AdjustXTODefaultValues;
  end;
end;

procedure TSerialPortNGXE.SetClusterSize(Value : Word);
begin
  FClusterSize := Value;
end;

procedure TSerialPortNGXE.SetRxQueueSize(Value : Word);
begin
  if Value <> FRxQueueSize then
  begin
    FRxQueueSize := Value;
    if not SetupComm(FHCommPort,FRxQueueSize,FTxQueueSize) then
      ProcessError(0101,GetLastError,'Error can not set RXQueueSize',enError);
  end;
end;

procedure TSerialPortNGXE.SetTxQueueSize(Value : Word);
begin
  if Value <> FTxQueueSize then
  begin
    FTxQueueSize := Value;
    if not SetupComm(FHCommPort,FRxQueueSize,FTxQueueSize) then
      ProcessError(0102,GetLastError,'Error can not set TXQueueSize',enError);
  end;
end;

procedure TSerialPortNGXE.SetErrorNoise(Value : Byte);
begin
  FErrorNoise := Value;
end;

procedure TSerialPortNGXE.SetReadRequest(Value : Boolean);
begin
  FReadRequest := Value;
end;

procedure TSerialPortNGXE.SetActive(NewState : Boolean);
begin
  // You may expect that this function set only the FActive Value
  // This is done by the PortWork procedure, depending from the successful
  // opened Port
  if NewState <> FActive then
    PortWork(NewState);
end;

//
// Several Methods

procedure TSerialPortNGXE.ProcessError(const Place, Code : Cardinal; const Msg : AnsiString; const Noise : Byte);
begin
  if ShutdownInProgress then Exit; // No Messages now the Component is in Destroystate
  if (Noise <= FErrorNoise) and
     Assigned(FOnProcessError) then
      FOnProcessError(Self,Place,Code,Msg,Noise); //Owner replaced by Self
end;

procedure TSerialPortNGXE.InitOverlapped(var Overlapped : TOverlapped);
begin
  Overlapped.Offset := 0;
  Overlapped.OffsetHigh := 0;
  Overlapped.Internal := 0;
  Overlapped.InternalHigh := 0;
  Overlapped.hEvent := CreateEvent(nil,True,False,'');
  if Overlapped.hEvent = 0 then
    ProcessError(1001,GetLastError,'Error Creating Overlapped Event',enError)
  else if GetLastError = ERROR_ALREADY_EXISTS then
    ProcessError(1002,ERROR_ALREADY_EXISTS,'Error Overlapped Event Exists',enError)
end;

procedure TSerialPortNGXE.ResetOverlapped(var Overlapped : TOverlapped);
begin
  if not ResetEvent(Overlapped.hEvent) then
    ProcessError(1101,GetLastError,'Error resetting Overlapped Event',enError);
end;

procedure TSerialPortNGXE.SetOverlapped(var Overlapped : TOverlapped);
begin
  if not SetEvent(Overlapped.hEvent) then
// EVENT_MODIFY_STATE
    ProcessError(1101,GetLastError,'Error resetting Overlapped Event',enError);
end;


//
// Create method.
constructor TSerialPortNGXE.Create(AOwner : TComponent);
var
  lVersionInfo : TOSVersionInfo;
begin
  inherited Create(AOwner);
  FReceiveDataClusterListLockCS := TCriticalSection.Create;
  ShutdownInProgress := False;
  FHCommPort := INVALID_HANDLE_VALUE;
  lVersionInfo.dwMajorVersion:= 0; // avoid FPC warning
  WinVersion := CheckOS(lVersionInfo);

  // Set initial settings.  Even though
  // the default parameter was specified
  // in the property, if you were to
  // create a component at runtime, the
  // defaults would not get set.  So it
  // is important to call them again in
  // the create of the component.
  FCommPort := dfltCommPort;
  FBaudRate := dfltBaudRate;
  FParityType := dfltParityType;
  FStopBits := dfltStopBits;
  FDataBits := dfltDataBits;
  FXONChar := dfltXONChar;
  FXOFFChar := dfltXOFFChar;
  FXONLimDiv := dfltXONLimDiv;
  FXOFFLimDiv := dfltXOFFLimDiv;
  FFlowControlFlags := dfltFlowControlFlags;
  FRTOCharDelayTime := dfltRTOCharDelayTime;
  FRTOExtraDelayTime := dfltRTOExtraDelayTime;
  FWTOCharDelayTime := dfltWTOCharDelayTime;
  FWTOExtraDelayTime := dfltWTOExtraDelayTime;
  FXTOAuto := dfltXTOAuto;
  FClusterSize := dfltClusterSize;
  FRxQueueSize := dfltRxQueueSize;
  FTxQueueSize := dfltTxQueueSize;
  FErrorNoise := enAll;
  FReadRequest := False;
  FRTSState := dfltRTSState;
  FDTRState := dfltDTRState;
  FBREAKState := dfltBREAKState;
  FThreadQuietMode := dfltThreadQuietMode;
  FAutoReadRequest := dfltAutoReadRequest;
  LastErr := 0;
  FReceiveDataClusterList := TObjectList.Create(True); // Create the List to store the received Clusters
  InitOverlapped(WriteOverlapped);
  InitOverlapped(ReadOverlapped);
  InitOverlapped(StatusOverlapped);
  WorkThread := TSerialPortNGXEWorkThread.Create(Self);
end;

// Destroy method.
destructor TSerialPortNGXE.Destroy;
begin
  ShutdownInProgress := True;
  PortWork(False);
  WorkThread.Terminate;
  WaitForThreadNotRunning();
  FileClose(WriteOverlapped.hEvent); { *Konvertiert von CloseHandle* }
  FileClose(StatusOverlapped.hEvent); { *Konvertiert von CloseHandle* }
  FileClose(ReadOverlapped.hEvent); { *Konvertiert von CloseHandle* }
  FReceiveDataClusterList.Clear;
  FReceiveDataClusterList.Free;
  WorkThread.WaitFor;
  WorkThread.Free;
  FReceiveDataClusterListLockCS.Free;
  inherited Destroy;
end;

// PortWork Closes or Opens the Port depending of the Parm
// It sets the fActive Variable depending of result of Opening the Port
procedure TSerialPortNGXE.PortWork(const ReOpen : Boolean);
var
  lasterr : Cardinal;
begin
  if FActive then // The Port is Open, Close first
  begin
    ProcessError(0100,0,'Msg start deactivating Port',enMsg);
    if not SetCommMask(FHCommPort,0) then
      ProcessError(0101,GetLastError,'Error disabling CommEvents',enError);
    fActive := False; // The WorkThread check this Flag
    if not PurgeComm(FHCommPort, PURGE_RXABORT or PURGE_RXCLEAR or PURGE_TXABORT or PURGE_TXCLEAR) then
      ProcessError(0102,GetLastError,'Error clearing Queues',enError);
    WaitForThreadNotRunning(); // Max 1000ms
    if WorkThreadIsRunning then
      ProcessError(0104,0,'Warning ThreadIsRunning',enWarning);
    SetSignalDTR(False);
    SetSignalRTS(False);
    FileClose(FHCommPort); { *Konvertiert von CloseHandle* }
//      ProcessError(0103,GetLastError,'Error closing Port',enError);
    FHCommPort := INVALID_HANDLE_VALUE;
  end;
// The Port is Closed, the Thread is Idle
  if ReOpen then
  begin // Reopen the Port with (new) Parms
    ProcessError(0110,0,Format('Msg start reactivating Port %s',[FCommPort]),enMsg);
    FHCommPort := CreateFile(PChar('\\.\'+FCommPort),
          GENERIC_READ OR GENERIC_WRITE,
          0,
          nil,
          OPEN_EXISTING,
          FILE_FLAG_OVERLAPPED,0);
    FActive := (FHCommPort <> INVALID_HANDLE_VALUE);
    if FActive then
    begin
      if not SetupComm(FHCommPort,FRxQueueSize,FTxQueueSize) then
        ProcessError(0111,GetLastError,'Error setup Queuesize',enError);
      SetupDCB;
      SetSignalDTR(FDTRState);
      SetSignalRTS(FRTSState);
      SetSignalBreak(FBreakState);
      EnableEvents;
    end
    else
    begin
      lasterr := GetLastError;
      ProcessError(0112,GetLastError,Format('Error reopening Port, LastError: %d (0x%4.4x)',[lasterr, lasterr]),enError);
    end;
  end;
end;

// Internal method to enable all Events
procedure TSerialPortNGXE.EnableEvents;
begin
  if not SetCommMask(FHCommPort, EV_BREAK or EV_CTS or EV_DSR or EV_ERR or EV_RING or EV_RLSD or EV_RXCHAR or EV_RXFLAG or EV_TXEMPTY) then
    ProcessError(0201,GetLastError,'Error activating CommEventMask',enError);
end;

// Public method to cancel and  flush the receive buffer.
procedure TSerialPortNGXE.ClearRxDQueue;
begin
  if FActive then
    if not PurgeComm(FHCommPort, PURGE_RXABORT or PURGE_RXCLEAR) then
      ProcessError(0301,GetLastError,'Error clearing RxD Queue',enError);
end;

// Public method to cancel and flush the transmit buffer.
procedure TSerialPortNGXE.ClearTxDQueue;
begin
  if FActive then
    if not PurgeComm(FHCommPort, PURGE_TXABORT or PURGE_TXCLEAR) then
      ProcessError(0401,GetLastError,'Error clearing TxD Queue',enError);
end;

// Public method to Play with the RTS Line
// It is an Error to work on this Line while in the Flowmode bmfOutxCtsFlow!
procedure TSerialPortNGXE.SetSignalRTS(State : Boolean);
begin
  if FActive then
  begin
    if State then
    begin
      if not EscapeCommFunction(FHCommPort,SETRTS) then
        ProcessError(0501,GetLastError,'Error setting RTS',enError)
    end
    else
    begin
      if not EscapeCommFunction(FHCommPort,CLRRTS) then
        ProcessError(0502,GetLastError,'Error clearing RTS',enError)
    end;
  end;
  FRTSState := State;
end;

// Public method to Play with the DTR Line
// It is an Error to work on this Line while in the Flowmode bmfOutxDtrFlow is set!
procedure TSerialPortNGXE.SetSignalDTR(State : Boolean);
begin
  if FActive then
  begin
    if State then
    begin
      if not EscapeCommFunction(FHCommPort,SETDTR) then
        ProcessError(0601,GetLastError,'Error setting DTR',enError)
    end
    else
    begin
      if not EscapeCommFunction(FHCommPort,CLRDTR) then
        ProcessError(0602,GetLastError,'Error clearing DTR',enError)
    end;
  end;
  FDTRState := State;
end;

// Public method to set the break State
procedure TSerialPortNGXE.SetSignalBREAK(State : Boolean);
begin
  if FActive then
  begin
    if State then
    begin
      if not SetCommBreak(FHCommPort) then
        ProcessError(0701,GetLastError,'Error setting BREAK State',enError)
    end
    else
    begin
      if not ClearCommBreak(FHCommPort) then
        ProcessError(0702,GetLastError,'Error clearing BREAK State',enError)
    end;
  end;
  FBREAKState := State;
end;

// Initialize the device control block.
procedure TSerialPortNGXE.SetupDCB;
var
  MyDCB : TDCB;
  MyCommTimeouts : TCommTimeouts;
//  SDCB : array[0..79] of AnsiChar;
  b : Byte;
begin
  // The GetCommState function fills in a
  // device-control block (a DCB structure)
  // with the current control settings for
  // a specified communications device.
  // (Win32 Developers Reference)
  // Get a default fill of the DCB.
  MyDCB.BaudRate:=0; //Avoid FPC warning
  if not GetCommState(FHCommPort, MyDCB) then
  begin
    ProcessError(0801,GetLastError,'Error getting DCB from CommState',enError);
    Exit;
  end;

  MyDCB.BaudRate := FBaudRate;
  MyDCB.Flags := bmfBinary; //Must be set under Win32, resetting all other Flags to '0'
  if FParityType <> ptNoParity then //NOPARITY then // If a ParityType is selceted, set Paritybit automatic
    MyDCB.Flags := MyDCB.Flags or bmfParity;
  case FParityType of
    ptNoParity    : b := NOPARITY;
    ptOddParity   : b := ODDPARITY;
    ptEvenParity  : b := EVENPARITY;
    ptMarkParity  : b := MARKPARITY;
    ptSpaceParity : b := SPACEPARITY;
  else
    b := NOPARITY;
  end;
  MyDCB.Parity := b;
  if FParityErrorReplacement then
    MyDCB.Flags := MyDCB.Flags or bmfErrorChar;

  // FloControl
  if fcOutxCtsFlow in FFlowControlFlags then // CTS output flow control
    MyDCB.Flags := MyDCB.Flags or bmfOutxCtsFlow;
  if fcOutxDsrFlow in FFlowControlFlags then // DSR output flow control
    MyDCB.Flags := MyDCB.Flags or bmfOutxDsrFlow;
  if fcDtrControlEnable in FFlowControlFlags then     // DTR Enable
    MyDCB.Flags := MyDCB.Flags or bmfDtrControlEnable;
  if fcDtrControlHandshake in FFlowControlFlags then  // DTR Handshake
    MyDCB.Flags := MyDCB.Flags or bmfDtrControlHandshake;
  if fcDsrSensitivity in FFlowControlFlags then       // DSR sensitivity
    MyDCB.Flags := MyDCB.Flags or bmfDsrSensitivity;
  if fcTXContinueOnXoff in FFlowControlFlags then   // XOFF continues Tx
    MyDCB.Flags := MyDCB.Flags or bmfTXContinueOnXoff;
  if fcOutX in FFlowControlFlags then               // XON/XOFF out flow control
    MyDCB.Flags := MyDCB.Flags or bmfOutX;
  if fcInX in FFlowControlFlags then                // XON/XOFF in flow control
    MyDCB.Flags := MyDCB.Flags or bmfInX;
  if fcRtsControlEnable in FFlowControlFlags then    // RTS Enable
    MyDCB.Flags := MyDCB.Flags or bmfRtsControlEnable;
  if fcRtsControlHandshake in FFlowControlFlags then // RTS Handshake
    MyDCB.Flags := MyDCB.Flags or bmfRtsControlHandshake;
  if fcRtsControlToggle  in FFlowControlFlags then    // RTS Toggle
    MyDCB.Flags := MyDCB.Flags or bmfRtsControlToggle;

  if FStripNullChars then
    MyDCB.Flags := MyDCB.Flags or bmfNull;
  MyDCB.ErrorChar := FParityErrorChar;
  MyDCB.EvtChar := FEventChar;
  case FStopBits of
    sbOneStopBit        : b := ONESTOPBIT;
    sbOnePtFiveStopBits : b := ONE5STOPBITS;
    sbTwoStopBits       : b := TWOSTOPBITS;
  else
    b := ONESTOPBIT;
  end;
  MyDCB.StopBits := b;
  MyDCB.ByteSize := FDataBits;
  MyDCB.XONChar := FXONChar;
  MyDCB.XOFFChar := FXOFFChar;
  MyDCB.XONLim := FRxQueueSize * FXONLimDiv div 100; // Send XOn if e.g fXONLimDiv = 33 -> 33% full
  MyDCB.XOFFLim := FRxQueueSize * FXOFFLimDiv div 100;  // Send XOff if e.g fXOffLimDiv = 33 -> 100%-33%=67% Percent full
  MyDCB.EOFChar := #0; //Ignored under Win32

  // The SetCommTimeouts function sets
  // the time-out parameters for all
  // read and write operations on a
  // specified communications device.
  // (Win32 Developers Reference)
  // The GetCommTimeouts function retrieves
  // the time-out parameters for all read
  // and write operations on a specified
  // communications device.
  MyCommTimeouts.ReadIntervalTimeout:=0; //avoid FPC warnings
  GetCommTimeouts(FHCommPort, MyCommTimeouts);
  //Read Timeouts are disabled here, because they realized manually in the WorkThread
  MycommTimeouts.ReadIntervalTimeout := MAXDWORD;
  MycommTimeouts.ReadTotalTimeoutMultiplier := 0;
  MycommTimeouts.ReadTotalTimeoutConstant := 0;
  //Write Timeouts disable here
  MycommTimeouts.WriteTotalTimeoutMultiplier := 0;
  MycommTimeouts.WriteTotalTimeoutConstant := 0;
  if not SetCommTimeouts(FHCommPort, MyCommTimeouts) then
      ProcessError(0802,GetLastError,'Error setting CommTimeout',enError);
  if not SetCommState(FHCommPort, MyDCB) then
    ProcessError(0802,GetLastError,'Error setting CommState, 87 indicate that Parms are incorrect',enError);
end;

// Public Send data method.
procedure TSerialPortNGXE.SendData(const Data; const Size : Cardinal); //Send binary Data
var MyCommTimeOuts : TCommTimeOuts;
begin
  if FSendInProgress then
  begin
    ProcessError(0901,0,'Msg, dont enter SendData while SendInProgress is set',enMsg);
    Exit;
  end;
  if Size = 0 then
  begin
    FSendInProgress := False;
    if Assigned(FOnWriteDone) then
      FOnWriteDone(Self);
    Exit;
  end;
  MyCommTimeouts.ReadIntervalTimeout:=0; //Avoid FPC warnings
  GetCommTimeouts(FHCommPort, MyCommTimeouts);
  //Read Timeouts are disabled
  MycommTimeouts.ReadIntervalTimeout := MAXDWORD;
  MycommTimeouts.ReadTotalTimeoutMultiplier := 0;
  MycommTimeouts.ReadTotalTimeoutConstant := 0;
  //Write Timeouts calculated from the settings
  MycommTimeouts.WriteTotalTimeoutMultiplier := 0;
  MycommTimeouts.WriteTotalTimeoutConstant := ((fWTOCharDelayTime*Size) div 1000) + fWTOExtraDelayTime;
  if not SetCommTimeouts(FHCommPort, MyCommTimeouts) then
      ProcessError(0902,GetLastError,'Error setting CommTimeout',enError);
  BytesToWrite := Size;
  if not WriteFile(FHCommPort,
            Data,
            Size,
            FWrittenBytes,
            @WriteOverlapped) then
  begin
    LastErr := GetLastError;
    if LastErr <> ERROR_IO_PENDING then
    begin
      ProcessError(0903,LastErr,'Error writing Data',enError);
      ResetOverlapped(WriteOverlapped);
      FSendInProgress := False;
    end
    else
    begin
      WriteStartTime := GetTickCount64;
      FSendInProgress := True;
    end;
   end
   else  // Write was done immidiatly
   begin
     if Assigned(FOnWriteDone) then
       FOnWriteDone(Self);
   end;
end;

// Public SendString Method
procedure TSerialPortNGXE.SendAnsiString(const S : AnsiString);
var
  b : Byte;
begin
  if Length(S) > 0 then
    SendData(S[1], Length(S))
  else
  begin
    b:=0;
    SendData(b,0);
  end;
end;
procedure TSerialPortNGXE.SendString(const S : String);
begin
  SendAnsiString(AnsiString(S));
end;
procedure TSerialPortNGXE.SendUnicodeString(const S : UnicodeString); //Send a Unicode String
var
  b : Byte;
begin
  if Length(S) > 0 then
    SendData(S[1], Length(S)*SizeOf(Char))
  else
  begin
    b := 0;
    SendData(b,0);
  end;
end;



// Public ExtractNextCluster Method
// Returns the next cluster object from the receiver queue.
// If no cluster available then Nil is returned.
// The caller has to free the passed Object.
function TSerialPortNGXE.ExtractNextCluster : TSerialNGXEDataCluster;
begin
  Result := Nil;
  FReceiveDataClusterListLockCS.Enter;
  try
    if FReceiveDataClusterList.Count > 0 then
    begin
      Result := TSerialNGXEDataCluster(FReceiveDataClusterList.Items[0]); //Element entnehmen
      FReceiveDataClusterList.Extract(Result);
    end;
  finally
    FReceiveDataClusterListLockCS.Leave;
  end;
end;
// Public NextClusterExists Method
// Returns True is a cluster is available
function TSerialPortNGXE.NextClusterExists : Boolean;
begin
  FReceiveDataClusterListLockCS.Enter;
  try
    Result := (FReceiveDataClusterList.Count > 0);
  finally
    FReceiveDataClusterListLockCS.Leave;
  end;
end;
// Public NextClusterSize Method
// Return the Number of Databytes
// 0..MAXINT indicates that a Cluster is available,
// 0 means, that no Data has been received, but an Error code exists
// -1 means no cluster is available
function TSerialPortNGXE.GetNextClusterSize : Integer;
begin
  Result := -1;
  FReceiveDataClusterListLockCS.Enter;
  try
    if FReceiveDataClusterList.Count > 0 then
      Result := TSerialNGXEDataCluster(FReceiveDataClusterList[0]).ClusterSize;
  finally
    FReceiveDataClusterListLockCS.Leave;
  end;
end;

// Public NextClusterCCError Method
// Returns the ErrorCode of the Next Cluster
// Returns MAXCardinal if no Cluster in List
function TSerialPortNGXE.GetNextClusterCCError : Cardinal;
begin
  Result := MAXDWORD;
  FReceiveDataClusterListLockCS.Enter;
  try
    if FReceiveDataClusterList.Count > 0 then
      Result := TSerialNGXEDataCluster(FReceiveDataClusterList[0]).ClusterCCError;
  finally
    FReceiveDataClusterListLockCS.Leave;
  end;
end;

// Public CopyNextClusterData Method
// Copy the data from the next cluster into the Data variable
// Returns the Number of bytes that was available in the cluster
// If Maxlength is smaller than the amount of data available in the cluster
// the rest of the data ist not copied, but lost
// It is a good idea to query the needed buffersize with GetNextClusterSize
// to adjust the size before calling this method
function TSerialPortNGXE.CopyNextClusterData(var Data; const MaxLength : Integer) : Integer;
var
  c : TSerialNGXEDataCluster;
  cnt : Integer;
begin
  Result := 0;
  c := ExtractNextCluster;
  try
    if Assigned(c) then
    begin
      Result := c.ClusterSize;
      cnt := Result;
      if cnt > MaxLength then
        cnt := MaxLength;
      if cnt > 0 then
        Move(c.FClusterData[0],Data,cnt);
    end;
  finally
    if Assigned(c) then
      c.Free;
  end;
end;
// Public ReadNextClusterAsAnsiString method
// The data from the next cluster is copied into the returned AnsiString
// All bytes are copied, including #0 or CR/LF etc.
// Use this function if you are expecting string like data
function TSerialPortNGXE.ReadNextClusterAsAnsiString : AnsiString;
var
  c : TSerialNGXEDataCluster;
begin
  Result := '';
  c := ExtractNextCluster;
  try
    if Assigned(c) then
    begin
      SetLength(Result,c.ClusterSize);
      if c.ClusterSize > 0 then
        Move(c.FClusterData[0],Result[1],c.ClusterSize);
    end;
  finally
    if Assigned(c) then
      c.Free;
  end;
end;
// Public ReadNextClusterAsString method
// same as ReadNextClusterAsAnsiString method
// The result is converted into a String
// Dont use this function if you expect Unicode data!
function TSerialPortNGXE.ReadNextClusterAsString : String;
var
  s : AnsiString;
begin
  s := ReadNextClusterAsAnsiString;
  Result := String(s);
end;
function TSerialPortNGXE.ReadNextClusterAsUnicodeString : UnicodeString;
var
  c : TSerialNGXEDataCluster;
  sz : Integer;
begin
  Result := '';
  c := ExtractNextCluster;
  try
    if Assigned(c) then
    begin
      sz := c.ClusterSize;
      if sz > 0 then
      begin
        if (SizeOf(Char) > 1) and Odd(sz) then
          SetLength(Result,(sz+1) div SizeOf(Char))
        else
          SetLength(Result,sz div SizeOf(Char));
        Move(c.FClusterData[0],Result[1],c.ClusterSize);
      end;
    end;
  finally
    if Assigned(c) then
      c.Free;
  end;
end;

// Public Method to fit the TimeOut Values to the current Baudrate
// If the Property XTOAuto is true this method will be called from the SetBaud method
procedure TSerialPortNGXE.AdjustXTODefaultValues;
var i : Integer;
    NewXTO : Cardinal;
begin
  NewXTO := 1100;
  for i := 0 to BaudRateCount-1 do
    begin
      if fBaudRate >= BaudRates[i] then
        NewXTO := XTOCharDelayDef[i];
    end;
  SetRTOCharDelayTime(NewXTO);
  SetWTOCharDelayTime(NewXTO);
end;

// Saves all Setting into the passed Strings
// e.g.: SerialPortNGXE.GetSettingsAsStrings(Memo1.Lines);
procedure TSerialPortNGXE.GetSettingsAsStrings(const Strings : TStrings);
begin
  if not Assigned(Strings) then Exit;
  Strings.Clear;
  Strings.Values['CommPort'] := FCommPort;
  Strings.Values['BaudRate'] := IntToStr(FBaudRate);
  Strings.Values['ParityType'] := IntToStr(Byte(FParityType));
  Strings.Values['ParityErrorChar'] := String(FParityErrorChar);
  Strings.Values['ParityErrorReplacement'] := IntToStr(Byte(FParityErrorReplacement));
  Strings.Values['StopBits'] := IntToStr(Byte(FStopBits));
  Strings.Values['DataBits'] := IntToStr(FDataBits);
  Strings.Values['XONChar'] := String(FXONChar);
  Strings.Values['XOFFChar'] := String(FXOFFChar);
  Strings.Values['XONLimDiv'] := IntToStr(FXONLimDiv);
  Strings.Values['XOFFLimDiv'] := IntToStr(FXOFFLimDiv);
  Strings.Values['FlowControl'] := IntToStr(Word(FFlowControlFlags));
  Strings.Values['StripNullChars'] := IntToStr(Byte(FStripNullChars));
  Strings.Values['EventChar'] := String(FEventChar);
  Strings.Values['RTOCharDelayTime'] := IntToStr(FRTOCharDelayTime);
  Strings.Values['RTOExtraDelayTime'] := IntToStr(FRTOExtraDelayTime);
  Strings.Values['ClusterSize'] := IntToStr(FClusterSize);
  Strings.Values['RxQueueSize'] := IntToStr(FRxQueueSize);
  Strings.Values['TxQueueSize'] := IntToStr(FTxQueueSize);
  Strings.Values['WTOCharDelayTime'] := IntToStr(FWTOCharDelayTime);
  Strings.Values['WTOExtraDelayTime'] := IntToStr(FWTOExtraDelayTime);
  Strings.Values['XTOAuto'] := IntToStr(Byte(FXTOAuto));
  Strings.Values['RTSState'] := IntToStr(Byte(FRTSState));
  Strings.Values['DTRState'] := IntToStr(Byte(FDTRState));
  Strings.Values['BREAKState'] := IntToStr(Byte(FBREAKState));
  Strings.Values['ErrorNoise'] := IntToStr(FErrorNoise);
  Strings.Values['ThreadQuietMode'] := IntToStr(Byte(FThreadQuietMode));
  Strings.Values['UseThreadSynchronize'] := IntToStr(Byte(FUseThreadSynchronize));
  Strings.Values['AutoReadRequest'] := IntToStr(Byte(FAutoReadRequest));
  Strings.Values['Active'] := IntToStr(Byte(FActive));
end;

// Sets all Settings from the passed Strings
// e.g.: SerialPortNGXE.WriteSettingsAsStrings(Memo1.Lines)
// an exception is created on parameter errors
procedure TSerialPortNGXE.WriteSettingsAsStrings(const Strings : TStrings);
var
  lActivate : Boolean;
    function CharFromStr(S : String) : Char;
    begin
      if Length(S) > 0 then
        Result := S[1]
      else
        Result := #0;
    end;

    function AnsiCharFromStr(S : String) : AnsiChar;
    begin
      if Length(S) > 0 then
        Result := AnsiChar(S[1])
      else
        Result := AnsiChar(#0);
    end;

    function StringFromStringsDef(const AName : String; const ADefValue : String) : String;
    var
      ndx : Integer;
    begin
      ndx := Strings.IndexOfName(AName);
      if ndx < 0 then
        Result := ADefValue
      else
        Result := Strings.ValueFromIndex[ndx];
    end;
    function IntFromStringsDef(const AName : String; const ADefValue : Integer) : Integer;
    var
      ndx : Integer;
    begin
      ndx := Strings.IndexOfName(AName);
      if ndx < 0 then
        Result := ADefValue
      else
        Result := StrToIntDef(Strings.ValueFromIndex[ndx],ADefValue);
    end;
    function BoolFromStringsDef(const AName : String; const ADefValue : Boolean) : Boolean;
    var
      ndx : Integer;
    begin
      ndx := Strings.IndexOfName(AName);
      if ndx < 0 then
        Result := ADefValue
      else
        Result := Strings.ValueFromIndex[ndx] = '1';
    end;


begin
  if not Assigned(Strings) then Exit;
  try
    lActivate := Active; //Save the current active state
    Active := False;
    if Length(Strings.Values['Active']) > 0 then //if the active state is defined
      lActivate := (Strings.Values['Active'] = '1'); //then use it
    CommPort := StringFromStringsDef('CommPort', FCommPort);
    FBaudRate := IntFromStringsDef('BaudRate',FBaudRate);
    FParityType := TSNXParityType(IntFromStringsDef('ParityType',Byte(FParityType)));
    ParityErrorChar := AnsiCharFromStr(StringFromStringsDef('ParityErrorChar',String(ParityErrorChar)));
    FParityErrorReplacement := BoolFromStringsDef('ParityErrorReplacement',FParityErrorReplacement);
    FStopBits := TSNXStopBits(IntFromStringsDef('StopBits', Byte(FStopBits)));
    FDataBits := IntFromStringsDef('DataBits', FDataBits);
    FXONChar := AnsiCharFromStr(StringFromStringsDef('XONChar', String(FXONChar)));
    FXOFFChar := AnsiCharFromStr(StringFromStringsDef('XOFFChar', String(FXOFFChar)));
    FXONLimDiv := IntFromStringsDef('XONLimDiv',FXONLimDiv);
    FXOFFLimDiv := IntFromStringsDef('XOFFLimDiv',FXOFFLimDiv);
//xxx    FFlowControlFlags := TSNXFlowControlFlags(IntFromStringsDef('FlowControl',Word(FFlowControlFlags)));
    FStripNullChars := BoolFromStringsDef('StripNullChars', FStripNullChars);
    FEventChar := AnsiCharFromStr(StringFromStringsDef('EventChar',String(FEventChar)));
    FRTOCharDelayTime := IntFromStringsDef('RTOCharDelayTime',FRTOCharDelayTime);
    FRTOExtraDelayTime := IntFromStringsDef('RTOExtraDelayTime',FRTOExtraDelayTime);
    FClusterSize := IntFromStringsDef('ClusterSize',FClusterSize);
    FRxQueueSize := IntFromStringsDef('RxQueueSize',FRxQueueSize);
    FTxQueueSize := IntFromStringsDef('TxQueueSize',FTxQueueSize);
    FWTOCharDelayTime := IntFromStringsDef('WTOCharDelayTime',FWTOCharDelayTime);
    FWTOExtraDelayTime := IntFromStringsDef('WTOExtraDelayTime',FWTOExtraDelayTime);
    FXTOAuto := BoolFromStringsDef('XTOAuto', FXTOAuto);
    FRTSState := BoolFromStringsDef('RTSState', FRTSState);
    FDTRState := BoolFromStringsDef('DTRState', FDTRState);
    FBREAKState := BoolFromStringsDef('BREAKState', FBREAKState);
    FErrorNoise := IntFromStringsDef('ErrorNoise',FErrorNoise);
    FThreadQuietMode := BoolFromStringsDef('ThreadQuietMode',FThreadQuietMode);
    FUseThreadSynchronize := BoolFromStringsDef('UseThreadSynchronize',FUseThreadSynchronize);
    FAutoReadRequest := BoolFromStringsDef('AutoReadRequest',FAutoReadRequest);

    Active := lActivate; //After all force the new settings
  except
    ProcessError(0402,0,'Error reading Settings',enError);
  end;
end;

procedure TSerialPortNGXE.WaitForThreadNotRunning(const WaitMS : Integer = 1000); //WaitMS waittime in Milliseconds
var
  cnt : Integer;
begin
  if WaitMS <= 0 then Exit;
  cnt := WaitMS div 10;
  while (cnt > 0) and (WorkThreadIsRunning) do
  begin
    Sleep(10);
    Dec(cnt);
  end;
end;

//
// WorkThread Definitions
// The Workthread manage all the Work in the Background
// - Checks wether the writing is done
// - Checks if Data are received
// - Checks the Status
// - Calls the Events

// Saves the process error Variables
procedure TSerialPortNGXEWorkThread.SetProcessError(const APlace, ACode : Cardinal; const AMsg : AnsiString; const ANoise : Byte);
begin
  Place := APlace;
  Code := ACode;
  Msg := AMsg;
  Noise := ANoise;
end;

// Calls the ProcessError Eventhandler
procedure TSerialPortNGXEWorkThread.ProcessError;
begin
  FOwner.ProcessError(Place,Code,Msg,Noise);
end;

// Create the Thread
constructor TSerialPortNGXEWorkThread.Create(AOwner : TSerialPortNGXE);
begin
  FOwner := AOwner;
  inherited Create(False);
end;

// Events...
procedure TSerialPortNGXEWorkThread.RxClusterEvent;
begin
  if Assigned(FOwner.FOnRxClusterEvent) then
    FOwner.FOnRxClusterEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.CommEvent;
begin
  if Assigned(FOwner.FOnCommEvent) then
    FOwner.FOnCommEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.CommStatEvent;
begin
  if Assigned(FOwner.FOnCommStat) then
    FOwner.FOnCommStat(FOwner);
end;
procedure TSerialPortNGXEWorkThread.BreakEvent;
begin
  if Assigned(FOwner.FOnBreakEvent) then
    FOwner.FOnBreakEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.CTSEvent;
begin
  if Assigned(FOwner.FOnCTSEvent) then
    FOwner.FOnCTSEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.DSREvent;
begin
  if Assigned(FOwner.FOnDSREvent) then
    FOwner.FOnDSREvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.LineErrorEvent;
begin
  if Assigned(FOwner.FOnLineErrorEvent) then
    FOwner.FOnLineErrorEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.RingEvent;
begin
  if Assigned(FOwner.FOnRingEvent) then
    FOwner.FOnRingEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.RIEvent;
begin
  if Assigned(FOwner.FOnRIEvent) then
    FOwner.FOnRIEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.RLSDEvent;
begin
  if Assigned(FOwner.FOnRLSDEvent) then
    FOwner.FOnRLSDEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.RxCharEvent;
begin
  if Assigned(FOwner.FOnRxCharEvent) then
    FOwner.FOnRxCharEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.RxEventCharEvent;
begin
  if Assigned(FOwner.FOnRxEventCharEvent) then
    FOwner.FOnRxEventCharEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.TxQueueEmptyEvent;
begin
  if Assigned(FOwner.FOnTxQueueEmptyEvent) then
    FOwner.FOnTxQueueEmptyEvent(FOwner);
end;
procedure TSerialPortNGXEWorkThread.WriteDone;
begin
  if Assigned(FOwner.FOnWriteDone) then
    FOwner.FOnWriteDone(FOwner);
end;

procedure TSerialPortNGXEWorkThread.ThreadSynchronize(Method: TThreadMethod);
begin
  if not FOwner.FThreadQuietMode then
    if FOwner.FUseThreadSynchronize
       // and (Application.Handle <> 0)
       // You can use this extra portion of security by the price to include the unit Forms
       then
      Synchronize(Method)
    else
      Method;
end;


//
// Workthread Maincycle
procedure TSerialPortNGXEWorkThread.Execute;
var
  WrittenBytes : Cardinal;
  BytesRead : Cardinal;
  CommStatus : TComStat;
  CommErrorCode : Cardinal;
  CommEventFlags : Cardinal;
  ModemState : Cardinal;
  RetCode : Cardinal;
  StartTime, t0 : Cardinal;
  Buffer : array of Byte;
  NeededBufferSize : Integer;
  WaitForReadEvent : Boolean;
  WaitForCommEvent : Boolean;
  HandleEvent : array[0..1] of Cardinal;
  ActiveMode, TerminateMode : Boolean;
  // The local Procedure evaluates the Events generated by the CommPort
  // and calles the Events of the Mainprogram
  procedure DoCommEvent;
  begin
    if FOwner.ShutdownInProgress then Exit;
    FOwner.FCommEvent := CommEventFlags;
    if (CommEventFlags and EV_BREAK) <> 0 then
      if Assigned(FOwner.FOnBreakEvent) then
        ThreadSynchronize(BreakEvent);
    if (CommEventFlags and EV_CTS) <> 0 then
      if Assigned(FOwner.FOnCTSEvent) then
        ThreadSynchronize(CTSEvent);
    if (CommEventFlags and EV_DSR) <> 0 then
      if Assigned(FOwner.FOnDSREvent) then
        ThreadSynchronize(DSREvent);
    if (CommEventFlags and EV_ERR) <> 0 then
      if Assigned(FOwner.FOnLineErrorEvent) then
        ThreadSynchronize(LineErrorEvent);
    if (CommEventFlags and EV_RING) <> 0 then
      if Assigned(FOwner.FOnRingEvent) then
        ThreadSynchronize(RingEvent);
    if (CommEventFlags and EV_RLSD) <> 0 then
      if Assigned(FOwner.FOnRLSDEvent) then
        ThreadSynchronize(RLSDEvent);
    if (CommEventFlags and EV_RXCHAR) <> 0 then
      if Assigned(FOwner.FOnRxCharEvent) then
        ThreadSynchronize(RxCharEvent);
    if (CommEventFlags and EV_RXFLAG) <> 0 then
      if FOwner.FAutoReadRequest then
        FOwner.ReadRequest := True;
    if Assigned(FOwner.FOnRxEventCharEvent) then
       ThreadSynchronize(RxEventCharEvent);
    if (CommEventFlags and EV_TXEMPTY) <> 0 then
      if Assigned(FOwner.FOnTxQueueEmptyEvent) then
        ThreadSynchronize(TxQueueEmptyEvent);
    if CommEventFlags <> 0 then
      if Assigned(FOwner.FOnCommEvent) then
         ThreadSynchronize(CommEvent);
  end;

  // Fetch the ModemStatus and CommErrorCode and CommStatus and generate
  // a CommStatEvent if something changed
  procedure GetStatus;
  var ExecDoCommEvent : Boolean;
      ExecRIEvent : Boolean;
      ClrCommErrDone : Boolean;
  begin
    ExecDoCommEvent := False;
    ExecRIEvent := False;
    if GetCommModemStatus(FOwner.FHCommPort,ModemState) then
    begin
      // There is a Bug in Win9x on signalizing the RING Event
      // We catch this manually here
      // The RingEvent is singnalize only on the falling edge of the RI!
      if FOwner.WinVersion = VER_PLATFORM_WIN32_WINDOWS then
      begin
        if ((ModemState and MS_RING_ON) = 0) and
           ((FOwner.FModemState and MS_RING_ON) <> 0) then
        // The RingIndicator Line has changed and is now False
        // generate Event
        begin
          CommEventFlags := EV_RING;
          FOwner.FRingState := (ModemState and MS_RING_ON) <> 0;
          ExecDoCommEvent := True;
        end;
      end;
      if ((ModemState xor FOwner.FModemState) and MS_RING_ON) <> 0 then
        ExecRIEvent := True;
      FOwner.FModemState := ModemState;
// Krystian from Poland suggest to add these 3 lines and got correct states
// even if no Event is assigned.
      FOwner.FCTSState := (ModemState and MS_CTS_ON) <> 0;
      FOwner.FDSRState := (ModemState and MS_DSR_ON) <> 0;
      FOwner.FRLSDState := (ModemState and MS_RLSD_ON) <> 0;
      FOwner.FRingState := (ModemState and MS_RING_ON) <> 0;
      if ExecRIEvent and Assigned(FOwner.FOnRIEvent) then
        ThreadSynchronize(RIEvent);
      if ExecDoCommEvent then
        DoCommEvent;
    end
    else
    begin
      SetProcessError(9905,GetLastError,'Error getting ModemStatus',enError);
      ThreadSynchronize(ProcessError);
    end;
    ClrCommErrDone := False;
    repeat
      if ClearCommError(FOwner.FHCommPort, CommErrorCode, @CommStatus) then
      begin
        if (FOwner.FCommError <> CommErrorCode) or
           (FOwner.FCommStateFlags <> CommStatus.flag0) or
           (FOwner.FCommStateInQueue <> CommStatus.cbInQue) or
           (FOwner.FCommStateOutQueue <> CommStatus.cbOutQue) then
        begin
          FOwner.FCommError := CommErrorCode;
          FOwner.FCommStateFlags := CommStatus.flag0;
          FOwner.FCommStateInQueue := CommStatus.cbInQue;
          FOwner.FCommStateOutQueue := CommStatus.cbOutQue;
          if Assigned(FOwner.FOnCommStat) then
            ThreadSynchronize(CommStatEvent);
        end
        else
          ClrCommErrDone := True;
      end
      else
      begin
        SetProcessError(9803,GetLastError,'Error ClearCommError',enError);
        ThreadSynchronize(ProcessError);
        ClrCommErrDone := True;
      end
    until ClrCommErrDone;
  end;

  // This local procedure checks if the Writing is done
  procedure CheckWriter;
  begin
    if FOwner.FSendInProgress then
    begin
      if GetOverlappedResult(FOwner.FHCommPort,FOwner.WriteOverlapped,WrittenBytes, FALSE) then
      begin
        FOwner.FWrittenBytes := WrittenBytes;
        FOwner.FSendInProgress := False;
        if WrittenBytes <> FOwner.BytesToWrite then
        begin
          SetProcessError(9701,RetCode,'Error write TimeOut',enError);
          ThreadSynchronize(ProcessError);
        end;
        ThreadSynchronize(WriteDone);
      end
      else
      begin
        RetCode := GetLastError;
        case RetCode of
          ERROR_IO_INCOMPLETE :;
          ERROR_IO_PENDING :
            begin
              t0 := GetTickCount64;
              if FOwner.WriteStartTime > t0 then
                FOwner.WriteStartTime := t0;
              if (t0-FOwner.WriteStartTime) > ((WrittenBytes*FOwner.FWTOCharDelayTime)/1000+FOwner.FWTOExtraDelayTime) then
              begin
                FOwner.FWrittenBytes := WrittenBytes;
                FOwner.FSendInProgress := False;
                FOwner.ResetOverlapped(FOwner.WriteOverlapped);
                SetProcessError(9701,RetCode,'Error write TimeOut',enError);
                ThreadSynchronize(ProcessError);
                ThreadSynchronize(WriteDone);
              end;
            end;
        else
          // Its an Error!!!
          FOwner.FSendInProgress := False;
          FOwner.ResetOverlapped(FOwner.WriteOverlapped);
          SetProcessError(9702,RetCode,'Error getting Overlapped Result',enError);
          ThreadSynchronize(ProcessError);
          ThreadSynchronize(WriteDone);
        end;
      end;
    end;
  end;

  //This procedure stores the received Cluster into the List
  procedure DoRxClusterStore;
  var
    c : TSerialNGXEDataCluster;
  begin
    if not FOwner.ShutdownInProgress then
    begin
      c := TSerialNGXEDataCluster.Create(Buffer[0],BytesRead,CommErrorCode);
      // The Storing of the Cluster into the Queue is done a CriticalSection
      FOwner.FReceiveDataClusterListLockCS.Enter;
      try
        try
          FOwner.FReceiveDataClusterList.Add(c);
          c := Nil;
        except
          if Assigned(c) then // If the cluster could not be stored in the queue, free cluster here
            c.Free;
        end;
      finally
        FOwner.FReceiveDataClusterListLockCS.Leave;
      end;
      ThreadSynchronize(RxClusterEvent);
    end;
  end;

  //Checks if Data is wainting in the RxDQueue and reads if Conditions are met
  //is called only if no Overlapp is running
  procedure ReadNoWait;
  begin
    if CommStatus.cbInQue = 0 then // No Char received
      StartTime := GetTickCount64 // Remember this Time as a Startpoint
    else // at least one Char was received
    begin
    // A Cluster is completed if one of the followoing conditions fit
    // - Owner request reading now
    // - cbInQue is greater than ClusterSize
    // - (cbInQue *  fRTOCharDelayTime)/1000 + fRTOExtraDelayTime is smaller than the elapsed Time
    // - a (Line-) Error occoured
      t0 := GetTickCount64;
      if FOwner.WriteStartTime > t0 then
        FOwner.WriteStartTime := t0;
      if (FOwner.FReadRequest) or
         (CommStatus.cbInQue >= FOwner.ClusterSize) or
         (( t0 - StartTime) > ((CommStatus.cbInQue *  FOwner.FRTOCharDelayTime)/1000 + FOwner.FRTOExtraDelayTime)) or
         ((CommErrorCode and (CE_RXOVER or CE_OVERRUN or CE_RXPARITY or CE_FRAME or CE_BREAK)) <> 0) then
      begin
        NeededBufferSize := CommStatus.cbInQue; //Ask the needed buffer size for this read
        if Length(Buffer) < NeededBufferSize then //increase the local buffer, if the buffer is smaller than needed
          SetLength(Buffer,NeededBufferSize);
        if ReadFile(FOwner.FHCommPort,
                    Buffer[0],
                    NeededBufferSize,
                    BytesRead,
                    @FOwner.ReadOverlapped) then
        begin //We have received something
          FOwner.FReadRequest := False; // Reset the Requestflag
          DoRxClusterStore; // Store Data and fire Event...
          StartTime := GetTickCount64; // Remember this Time as a Startpoint
        end
        else  // ReadFile was not successful, this may caused by the Overlapped function
        begin
          RetCode := GetLastError;
          if RetCode = ERROR_IO_PENDING then // Yes, Reading is in Progress
            WaitForReadEvent := True
          else
          begin // Error while reading
            FOwner.FReadRequest := False;
            SetProcessError(9804,RetCode,'Error reading Data',enError);
            ThreadSynchronize(ProcessError);
          end;
        end;
      end;
    end;
  end;

  // Checks for new events
  //is called only if no Overlapp is running
  procedure CommEventNoWait;
  begin
    CommEventFlags := 0;
    if WaitCommEvent(FOwner.FHCommPort,CommEventFlags,@FOwner.StatusOverlapped) then
    begin
      GetStatus;  // Update Statusflags 25.3.2003
      DoCommEvent; // Event Occours, fire Events
    end
    else
    begin
      RetCode := GetLastError;
      if RetCode = ERROR_IO_PENDING then
        WaitForCommEvent := True //Check the Overlapped.hEvent
      else
      begin
        SetProcessError(9907,RetCode,'Error calling WaitCommEvent',enError);
        ThreadSynchronize(ProcessError);
      end;
    end;
  end;

  // Checks for received Data while an Overlapp is running
  procedure ProcessWaitForRead;
  begin
    if not GetOverlappedResult(FOwner.FHCommPort,FOwner.ReadOverlapped,BytesRead, False) then
    begin
      RetCode := GetLastError;
      if RetCode = ERROR_OPERATION_ABORTED then
        SetProcessError(9907,RetCode,'Error read aborted',enError)
      else
        SetProcessError(9908,RetCode,'Error getting Overlappedresult',enError);
      ThreadSynchronize(ProcessError);
    end
    else  // Successfull Overlapped read
    begin
      DoRxClusterStore; // Store Data and fire Event...
      StartTime := GetTickCount64; // Remember this Time as a Startpoint
    end;
    WaitForReadEvent := False;
  end;

  // Checks for new Events while an Overlapp is running
  procedure ProcessWaitForComm;
  begin
    if (FOwner.FActive) then
    begin
      GetStatus;
      DoCommEvent;
    end;
    WaitForCommEvent := False;
  end;


// Main Cycle of the Thread
begin
  try
    SetLength(Buffer,1024); //Default Buffer 1K-Byte, will be increased as needed,
    StartTime := GetTickCount64;
    WaitForCommEvent := False;
    WaitForReadEvent := False;
    ActiveMode := FOwner.FActive;
    TerminateMode := Terminated;
    while not TerminateMode do
    begin
      try
        if ActiveMode then
        begin
          FOwner.WorkThreadIsRunning := True;
          if (FOwner.FActive) then
            GetStatus; // Picup several Information about the actual State of Com
          CheckWriter; // Checks for pending Writeprocess
          if (not WaitForReadEvent) and (FOwner.FActive) then // Start new Action only if not deactivating
            ReadNoWait; // Reads if avail, no waiting here
          if not WaitForCommEvent and (FOwner.FActive) then   // Start new Action only if not deactivating
            CommEventNoWait; // Check for Events, no waiting here
          // WaitForMultiple Events
          if (WaitForReadEvent and WaitForCommEvent) then
          begin
            HandleEvent[0] := FOwner.ReadOverlapped.hEvent;
            HandleEvent[1] := FOwner.StatusOverlapped.hEvent;
            RetCode := WaitForMultipleObjects(2,@HandleEvent,False,75);
            if (FOwner.FActive) then
              GetStatus; // Picup several Information about the actual State of Com
            case RetCode of
              WAIT_OBJECT_0 :
                begin
                  ProcessWaitForRead;
                end;
              WAIT_OBJECT_0 + 1 :
                begin
                  ProcessWaitForComm;
                end;
              WAIT_TIMEOUT :
                begin
                end;
            else
              SetProcessError(9911,RetCode,'Error getting Overlappedresult',enError);
              ThreadSynchronize(ProcessError);
              WaitForReadEvent := False;
              WaitForCommEvent := False;
            end;
          end
          else if WaitForReadEvent then
          begin
            RetCode := WaitForSingleObject(FOwner.ReadOverlapped.hEvent,75);
            if (FOwner.FActive) then
              GetStatus; // Picup several Information about the actual State of Com
            case RetCode of
              WAIT_OBJECT_0 :
                begin
                  if (FOwner.FActive) then
                    ProcessWaitForRead;
                end;
              WAIT_TIMEOUT :
                begin
                end;
            else
              SetProcessError(9912,RetCode,'Error getting Overlappedresult',enError);
              ThreadSynchronize(ProcessError);
              WaitForReadEvent := False;
            end;
          end
          else if WaitForCommEvent then// WaitForCommEvent
          begin
            RetCode := WaitForSingleObject(FOwner.StatusOverlapped.hEvent,75);
            if (FOwner.FActive) then
              GetStatus; // Picup several Information about the actual State of Com
            case RetCode of
              WAIT_OBJECT_0 :
                begin
                  ProcessWaitForComm;
                end;
              WAIT_TIMEOUT :
                begin
                end;
            else
              SetProcessError(9913,RetCode,'Error getting Overlappedresult',enError);
              ThreadSynchronize(ProcessError);
              WaitForCommEvent := False;
            end;
          end;
          if not FOwner.FActive then // The owner wants to deactivate the port
          begin
            if WaitForReadEvent then
              FOwner.SetOverlapped(FOwner.ReadOverlapped);
            if WaitForCommEvent then
              begin
                SetCommMask(FOwner.FHCommPort,0);
                FOwner.SetOverlapped(FOwner.StatusOverlapped);
              end;
            if (not WaitForCommEvent) and
               (not WaitForReadEvent) and
               (not FOwner.FSendInProgress) then
              ActiveMode := False; // We do so if everything pending is finished
          end;
        end
        else // not ActiveMode
        begin
          FOwner.WorkThreadIsRunning := False;
          ActiveMode := FOwner.FActive;
          Sleep(200); //prevent consuming 100% of cpu-time in inactive mode
        end;
        if Terminated and (not ActiveMode) then //No termination before deactivation
          TerminateMode := True;
      except
        on E : Exception do
          SetProcessError(9914,0,'Unhandled Exception',enError);
      end;
    end; // while not Terminated
  finally
    if not Terminated then
      Terminate;
    if Assigned(FOwner) then
    begin
      FOwner.WorkThreadIsTerminated := True;
      FOwner.WorkThreadIsRunning := False;
    end;
  end;
end;

end.
